Package: maaslin3
Title: "Refining and extending generalized multivariate linear models
        for meta-omic association discovery"
Year: 2025
Version: 1.4.0
Authors@R: c(
  person("William", "Nickols", email = "willnickols@g.harvard.edu", role = c("aut", "cre"), comment=c(ORCID="0000-0001-8214-9746")), 
  person("Jacob", "Nearing", email = "nearing@broadinstitute.org", role = "aut")
  )
Depends: R (>= 4.4)
Description: MaAsLin 3 refines and extends generalized multivariate
        linear models for meta-omicron association discovery. It finds
        abundance and prevalence associations between microbiome
        meta-omics features and complex metadata in population-scale
        epidemiological studies. The software includes multiple
        analysis methods (including support for multiple covariates,
        repeated measures, and ordered predictors), filtering,
        normalization, and transform options to customize analysis for
        your specific study.
License: MIT + file LICENSE
Imports: data.table, collapse, lmerTest, lme4, optparse, logging,
        mirai, multcomp, ggplot2, RColorBrewer, patchwork, scales,
        rlang, ggnewscale, survival, methods, BiocGenerics, reformulas
Suggests: knitr, testthat (>= 2.1.0), rmarkdown, markdown, kableExtra,
        dplyr, SummarizedExperiment, TreeSummarizedExperiment
VignetteBuilder: knitr
Collate: fit.R utility_scripts.R viz.R maaslin3.R
URL: http://huttenhower.sph.harvard.edu/maaslin3
biocViews: Metagenomics, Software, Microbiome, Normalization,
        MultipleComparison
BugReports: https://github.com/biobakery/maaslin3/issues
NeedsCompilation: no
Packaged: 2026-04-29 23:18:07 UTC; root
RoxygenNote: 7.3.3
Config/pak/sysreqs: cmake make
Repository: https://bioc-release.r-universe.dev
Date/Publication: 2026-04-28 13:04:13 UTC
RemoteUrl: https://github.com/bioc/maaslin3
RemoteRef: RELEASE_3_23
RemoteSha: f7a411ede14c49c6173799d252d6896bbae95450
Author: William Nickols [aut, cre] (ORCID:
    <https://orcid.org/0000-0001-8214-9746>),
  Jacob Nearing [aut]
Maintainer: William Nickols <willnickols@g.harvard.edu>
Built: R 4.6.0; ; 2026-04-30 00:35:31 UTC; unix
