export(maaslin3)
export(maaslin_check_arguments)
export(maaslin_log_arguments)
export(maaslin_read_data)
export(maaslin_reorder_data)
export(maaslin_compute_formula)
export(maaslin_check_formula)
export(maaslin_filter)
export(maaslin_process_metadata)
export(maaslin_normalize)
export(maaslin_transform)
export(maaslin_fit)
export(maaslin_write_results)
export(maaslin_write_results_lefse_format)
export(maaslin_plot_results)
export(maaslin_plot_results_from_output)
export(maaslin_contrast_test)
export(preprocess_dna_mtx)
export(preprocess_taxa_mtx)

import("optparse")
importFrom("stats", "as.formula", "coef", "glm", "lm", "na.exclude",
    "p.adjust", "update", "relevel", "anova", "approx", "filter", "median",
    "model.matrix", "pbeta", "quantile", "terms",
    "update.formula", "var", "vcov", "rnorm", "cov", "setNames")
importFrom("utils", "capture.output", "read.table", "write.table", "tail")
importFrom("stats", "sd", "na.omit")
importFrom("utils","type.convert")
importFrom("rlang", ".data", "check_installed")
importFrom("methods", "is")
importFrom("survival", "strata", "coxph", "Surv")
